/*
Scanner for C-Source TeX Formatter
(c) 1995 Martin Mares \& Robert Spalek
*/

#include "cstf.h"

#define STRBUFSIZE 8192

int class,value,line;
byte strbuf[STRBUFSIZE];

static byte linebuf[256],*curr;
static int processing_directive;

void err(char *c)
{
	fprintf(stderr,"Error in line %d: %s !\n",line,c);
	exit(FAIL_RETURN_CODE);
}

/* Character table */

static byte Chrt[256]={
    0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 1, 0, 3, 3, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
/*%   !  "  #  $  %  &  '  (  )  *  +  ,  -  .  /  */
    4,70, 5, 6,18,71,72, 7, 8, 9,73,74,41,75,40,76,
/*%0  1  2  3  4  5  6  7  8  9  :  ;  <  =  >  ?  */
   10,10,10,10,10,10,10,10,10,10,42,11,77,78,79,43,
/*%@  A  B  C  D  E  F  G  H  I  J  K  L  M  N  O  */
   18,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,
/*%P  Q  R  S  T  U  V  W  X  Y  Z  [  \  ]  ^  _  */
   12,12,12,12,12,12,12,12,12,12,12,13,15,14,80,12,
/*%`  a  b  c  d  e  f  g  h  i  j  k  l  m  n  o  */
   18,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,
/*%p  q  r  s  t  u  v  w  x  y  z  {  |  }  ~  7F */
   12,12,12,12,12,12,12,12,12,12,12,16,81,17,44, 0,
   18,18,18,18,18,18,18,18,18,18,18,18,18,18,18,18,
   18,18,18,18,18,18,18,18,18,18,18,18,18,18,18,18,
   18,18,18,18,18,18,18,18,18,18,18,18,18,18,18,18,
   18,18,18,18,18,18,18,18,18,18,18,18,18,18,18,18,
   18,18,18,18,18,18,18,18,18,18,18,18,18,18,18,18,
   18,18,18,18,18,18,18,18,18,18,18,18,18,18,18,18,
   18,18,18,18,18,18,18,18,18,18,18,18,18,18,18,18,
   18,18,18,18,18,18,18,18,18,18,18,18,18,18,18,18
};

#define CATE_INVALID 0
#define CATE_EOL 1
#define CATE_TAB 2
#define CATE_IGNORED 3
#define CATE_SPACE 4
#define CATE_QUOTE 5
#define CATE_HASH 6
#define CATE_APOST 7
#define CATE_LPAREN 8
#define CATE_RPAREN 9
#define CATE_DIGIT 10
#define CATE_SEMIC 11
#define CATE_LETTER 12
#define CATE_LBRACK 13
#define CATE_RBRACK 14
#define CATE_BACKSLASH 15
#define CATE_LGROUP 16
#define CATE_RGROUP 17
#define CATE_OTHER 18
#define CATE_FIRSTOP 40
#define CATE_FIRSTSTATE 70

/* State machine for operator recognition */

static byte Machine[] = {
/*  0 ! */		'!', 105, 12,
/*  1 % */		'=', 106,118,
/*  2 & */		'&',  13, 14,
/*  3 * */		'=', 108,122,
/*  4 + */		'+', 109, 15,
/*  5 - */		'-', 110, 16,
/*  6 / */		'/', 200, 18,
/*  7 < */		'=', 111, 20,
/*  8 = */		'=', 112,133,
/*  9 > */		'=', 113, 22,
/* 10 ^ */		'=', 114,137,
/* 11 | */		'|',  24, 25,
/* 12 ! */		'=', 116,117,
/* 13 && */		'=', 119,107,
/* 14 & */		'=', 120,121,
/* 15 + */		'=', 123,124,
/* 16 - */		'=', 125, 17,
/* 17 - */		'>', 126,127,
/* 18 / */		'*', 201, 19,
/* 19 / */		'=', 128,129,
/* 20 < */		'<',  21,130,
/* 21 << */		'=', 131,132,
/* 22 > */		'>',  23,134,
/* 23 >> */		'=', 135,136,
/* 24 || */		'=', 138,115,
/* 25 | */		'=', 139,140
};

static int readline(void)
{
	if (feof(infile)) { curr=0; class=L_EOF; return 1; }
	curr=linebuf;
	line++;
	if (!fgets(linebuf,256,infile)) { curr=0; class=L_EOF; return 1; }
	if (!strchr(curr,'\n')) err("line too long or not terminated");
	return 0;
}

static void gather_ident(void)
{
byte *t=strbuf;
	do
		{
		*t++ = *curr++;
		}
	while (Chrt[*curr] == CATE_LETTER || Chrt[*curr] == CATE_DIGIT);
	*t=0;
}

void parse(void)
{
byte cate;

	if (processing_directive)
		if (Chrt[*curr] != CATE_SPACE && Chrt[*curr] != CATE_TAB)
			{
			processing_directive = 0;
			if (*curr == '<')
				{
				byte *t = strbuf;
					while (*++curr != '>')
						switch (Chrt[*curr])
							{
							case CATE_INVALID:
							case CATE_EOL:
							case CATE_TAB:
							case CATE_SPACE:
							case CATE_IGNORED:
								err("Inconsistent directive argument");
							default:
								*t++ = *curr;
							}
					*t = 0;
					curr++;
					if (!*strbuf) err("Empty argument of #include");
					class = L_INCLARG;
					return;
				}
			}

	if (!curr || !*curr)
		{
		int sp;
			if (readline()) return;

			if (Chrt[*curr] == CATE_HASH)
				{
				curr++;
				gather_ident();
				processing_directive = 1;
				class = L_PREPROC;
				return;
				}

			sp=0;
			for(;;)
				{
				switch (Chrt[*curr++])
					{
					case CATE_SPACE:
						sp++;
						continue;
					case CATE_TAB:
						sp = TabSize + sp - (sp % TabSize);
						/* Fall-thru ! */
					case CATE_IGNORED:
						continue;
					}
				curr--;
				break;
				}
			if (sp)
				{
				class = L_SPACE;
				value = sp;
				return;
				}
		}

again:
	cate = Chrt[*curr++];
	switch (cate)
		{
		case CATE_INVALID:
		case CATE_OTHER:
			err("Invalid character encountered");
		case CATE_EOL:
			class = L_EOL;
			return;
		case CATE_TAB:
		case CATE_SPACE:
			while (Chrt[*curr] == CATE_TAB || Chrt[*curr] == CATE_SPACE) curr++;
			if (Chrt[*curr] == CATE_EOL) goto again;
			class = L_SPACE;
			value = 0;
			return;
		case CATE_IGNORED:
			goto again;
		case CATE_QUOTE:
			{
			byte *t = strbuf;
				for(;;)
					{
					if (t > strbuf+STRBUFSIZE-3)
						err("String constant too long");
					switch (Chrt[*curr++])
						{
						case CATE_QUOTE:
							*t = 0;
							class = L_STRING;
							return;
						case CATE_INVALID:
							err("Invalid character in string");
						case CATE_EOL:
							err("String constant not terminated");
						case CATE_TAB:
							*t++ = '\\';
							*t++ = 't';
							/* Fall-thru */
						case CATE_IGNORED:
							break;
						case CATE_BACKSLASH:
							*t++ = curr[-1];
							*t++ = *curr;
							if (*curr++ == '\n')
								{
								if (readline())
									err("EOF in the middle of string constant");
								}
							break;
						default:
							*t++ = curr[-1];
						}
					}
			}
		case CATE_HASH:
			class = L_OPERATOR;
			if (*curr == '#')
				{
				curr++;
				value = OP_HASHHASH;
				}
			else value = OP_HASH;
			return;
		case CATE_APOST:
			{
			byte *t = strbuf;
				for(;;)
					{
					if (t > strbuf+5)
						err("Character constant too long");
					switch (Chrt[*curr++])
						{
						case CATE_APOST:
							if (t == strbuf)
								err("Empty character constant");
							*t = 0;
							class = L_CHAR;
							return;
						case CATE_INVALID:
							err("Invalid character in char constant");
						case CATE_TAB:
							*t++ = '\\';
							*t++ = 't';
							/* Fall-thru */
						case CATE_IGNORED:
							break;
						case CATE_BACKSLASH:
							*t++ = curr[-1];
							*t++ = *curr;
							if (*curr++ != '\n') break;
							/* Fall-thru */
						case CATE_EOL:
							err("Character constant exceeds line");
						default:
							*t++ = curr[-1];
						}
					}
			}
		case CATE_DIGIT:
			{
			byte *t=strbuf;
				*t++ = curr[-1];
				if ((*curr == 'x' || *curr == 'X') && t[-1] == '0')
					{
					do
						{
						*t++ = *curr++;
						}
					while (*curr >= '0' && *curr <= '9'
						||	 *curr >= 'A' && *curr <= 'F'
						||	 *curr >= 'a' && *curr <= 'f');
					}
				else for(;;)
					{
					if (*curr >= '0' && *curr <= '9' || *curr == '.') *t++ = *curr++;
					else if (*curr == 'e' || *curr == 'E')
						{
						*t++ = *curr++;
						if (*curr == '+' || *curr == '-') *t++ = *curr++;
						}
					else break;
					}
				*t=0;
				class = L_NUMBER;
				return;
			}
		case CATE_LETTER:
			{
			struct symrec *s;
				curr--;
				gather_ident();
				s = id_lookup(strbuf);
				if (s) class=s->cate;
				else class=L_WORD;
				return;
			}

		case CATE_BACKSLASH:
			if (Chrt[*curr] != CATE_EOL)
				err("Backslash not allowed here");
			/* Fall-thru */

		case CATE_LPAREN:
		case CATE_RPAREN:
		case CATE_SEMIC:
		case CATE_LBRACK:
		case CATE_RBRACK:
		case CATE_LGROUP:
		case CATE_RGROUP:
			value = cate;
			class = L_SPECIAL;
			return;
		}

		if (cate >= CATE_FIRSTOP && cate < CATE_FIRSTSTATE)
			{
			class = L_OPERATOR;
			value = cate - CATE_FIRSTOP;
			return;
			}

		if (cate >= CATE_FIRSTSTATE)
			{
			int state = (cate - CATE_FIRSTSTATE);
				for(;;)
					{
					state *= 3;
					if (Machine[state++] == *curr) curr++;
					else state++;
					state = Machine[state];
					if (state >= 200)
						{
						int startline = (curr == linebuf+2);
						byte *t = strbuf;
							class = L_COMMENT;
							if (state == 200)		/* C++ Comment */
								while (*curr != '\n') *t++ = *curr++;
							else						/* C Comment */
								{
								while (*curr != '*' || curr[1] != '/')
									{
									*t++ = *curr;
									if (*curr++ == '\n')
										{
										class = L_SPLITCOMMENT;
										if (readline())
											err("Comment not terminated");
										}
									}
								curr += 2;
								}
							if (startline)
								class = L_FULLCOMMENT;
						*t = 0;
						if (*strbuf == '#')		/* Processing options */
							{
							byte c,*k,*z=strbuf+1;
								for(;;)
									{
									while (*z && *z <= ' ') z++;
									if (!*z) goto again;
									k=z;
									while (*z > ' ') z++;
									c=*z; *z=0;
									process_option(k);
									*z=c;
									}
							}
						return;
						}
					if (state >= 100)
						{
						class = L_OPERATOR;
						value = state-100;
						return;
						}
					}
			}
}

/* Keyword table */

static byte kwdinit[] = "\
volatile\000\020\
extern\000\020\
static\000\020\
register\000\020\
auto\000\020\
const\000\020\
short\000\013\
long\000\013\
float\000\013\
double\000\013\
void\000\013\
unsigned\000\013\
signed\000\013\
char\000\013\
int\000\013\
struct\000\013\
union\000\013\
enum\000\013\
sizeof\000\021\
typedef\000\012\
return\000\012\
continue\000\012\
if\000\012\
else\000\012\
break\000\012\
for\000\012\
do\000\012\
while\000\012\
switch\000\012\
case\000\012\
default\000\012\
goto\000\012\
\000";

#define HASH_CAPACITY 256
#define HASH_PRIME 101

static struct symrec *hash_table[HASH_CAPACITY];

static unsigned int calc_hash(byte *n)
{
unsigned int k = strlen(n);
	while (*n)
		k = HASH_PRIME*k + *n++;
	return k & (HASH_CAPACITY-1);
}

static struct symrec *id_lookup2(byte *name,unsigned int hash)
{
struct symrec *r = hash_table[hash];
	while (r)
		{
		if (!strcmp(r->name,name)) break;
		r=r->next;
		}
	return r;
}

struct symrec *id_lookup(byte *name)
{
	return id_lookup2(name,calc_hash(name));
}

struct symrec *new_symbol(byte *name)
{
struct symrec *r;
int hash = calc_hash(name);
	if (!(r=id_lookup2(name,hash)))
		{
		r = xmalloc(sizeof(struct symrec) + strlen(name));
		strcpy(r->name,name);
		r->next = hash_table[hash];
		hash_table[hash] = r;
		r->cate = L_WORD;
		}
	return r;
}

void init_keywords(void)
{
byte *z = kwdinit;
	while (*z)
		{
		struct symrec *sr = new_symbol(z);
			while (*z++) ;
			sr->cate = *z++;
		}
}
