/*
C-Source TeX Formatter
(c) 1995 Martin Mares \& Robert Spalek
*/

#include "cstf.h"

FILE *infile,*outfile;

void Err(char *c)
{
	fprintf(stderr,"%s !\n",c);
	exit(FAIL_RETURN_CODE);
}

void *xmalloc(int size)
{
void *p = malloc(size);
	if (!p) Err("Insufficient memory");
	return p;
}

byte SourceName[256],DestName[256];

static byte *GetExtension(byte *c)
{
byte *f=NULL;
	while (*c)
		switch (*c++)
			{
			case '/':
				f=NULL;
				break;
			case '.':
				f=c;
				break;
			}
	return f;
}

static void ExtendName(byte *c,byte *e)
{
byte *l = GetExtension(c);
	if (!l)
		{
		l=c;
		while (*l) l++;
		*l++ = '.';
		strcpy(l,e);
		}
}

static void SetExtension(byte *c,byte *e)
{
byte *l = GetExtension(c);
	if (!l) ExtendName(c,e);
	else strcpy(l,e);
}

int main(int argc,char **argv)
{
int fna=0;

	init_keywords();

	if (argc == 2 && ( !strcmp(argv[1],"--help") || !strcmp(argv[1],"?") ))
		{
		printf("\
C-Source TeX Formatter 0.3      (c) 1998 Martin Mares & Robert Spalek\n\n\
Usage:   cstf [<options>] <source> [<destination>]\n\n\
Options: -tabsize=<n>     Size of one tab in C source\n\
         -ssv,-ssi,-ssa   String spaces: visible, invisible, auto\n\
         -bye             Insert \\bye to TeX source\n\
         -nomac           Don't prepend \\input cstfmac.tex\n\
         -cmd=<s>         Typeset <s> as a command\n\
         -type=<s>        Typeset <s> as a type\n\
         -func=<s>        Typeset <s> as a function\n\
         -class=<s>       Typeset <s> as a storage class\n\
         -debug=<n>       Set debugging level\n\
");
		exit(0);
		}
	while (argc > 1)
		{
		argc--; argv++;
		if (strlen(*argv) > 248)
			Err("Command line argument too long");
		if (**argv == '-')
			process_option(*argv+1);
		else
			switch (fna++)
				{
				case 0:
					strcpy(SourceName,*argv);
					ExtendName(SourceName,"c");
					break;
				case 1:
					strcpy(DestName,*argv);
					ExtendName(DestName,"tex");
					break;
				case 2:
					Err("Too many file names");
				}
		}

	if (fna == 1)
		{
		strcpy(DestName,SourceName);
		SetExtension(DestName,"tex");
		}
	if (fna)
		{
		printf("Converting `%s' to `%s'...\n",SourceName,DestName);
		if (! (infile = fopen(SourceName,"r")))
			Err("Unable to open source file");
		if (! (outfile = fopen(DestName,"w")))
			Err("Unable to open destination file");
		}
	else
		{
		infile=stdin;
		outfile=stdout;
		}

	init_list();
	init_tex();

	for(;;)
		{
		create_list();
		if (! firstnode.next->next) break;
		if (debug_value & DEBUG_SCAN) debug_list();
		process_list();
		if (debug_value & DEBUG_PROC) debug_list();
		output_tex();
		delete_list();
		}

	end_tex();
	return 0;
}
