/*
C-Source TeX Formatter -- Global Declarations
(c) 1995 Martin Mares & Robert Spalek
*/

#define FAIL_RETURN_CODE 1

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

/* Types */

typedef unsigned char byte;

/* Defined in cstf.c */

void Err(char *);
void *xmalloc(int);
extern FILE *infile,*outfile;

/* Defined in options.c */

void process_option(byte *);

extern int TabSize,StringMode,InsertBye,InsertMacros,debug_value;

#define DEBUG_SCAN 1
#define DEBUG_PROC 2

/* Defined in clex.c */

struct symrec {
	struct symrec *next;
	byte cate;
	byte name[1];
	};

void err(char *);
void parse(void);
void init_keywords(void);
struct symrec *id_lookup(byte *);
struct symrec *new_symbol(byte *);

extern int class,value,line;
extern byte strbuf[];

/* Defined in lists.c */

struct listnode {
	struct listnode *next,*prev;
	byte cate,flag;
	unsigned int val;
	};

#define LVAL(x) ((x)->val)
#define LSTR(x) ((byte *) (x)->val)

extern struct listnode firstnode,lastnode;

void init_list(void);
void create_list(void);
void delete_list(void);
void debug_list(void);
void process_list(void);

/* Defined in tex.c */

void init_tex(void);
void output_tex(void);
void end_tex(void);

/* Lexical classes */

#define L_EOF 0
#define L_COMMENT 1
#define L_FULLCOMMENT 2
#define L_OPERATOR 3
#define L_CHAR 4
#define L_NUMBER 5
#define L_STRING 6
#define L_EOL 7
#define L_INCLARG 8
#define L_PREPROC 9
#define L_COMMAND 10
#define L_TYPE 11
#define L_WORD 12
#define L_SPECIAL 13
#define L_SPACE 14
#define L_SPLITCOMMENT 15
#define L_STORCLASS 16
#define L_FUNCTION 17
#define L_COMMENTFILL 18
#define L_VAR 19

/* Operator codes */

#define OP_DOT 0
#define OP_COMMA 1
#define OP_COLON 2
#define OP_QUEST 3
#define OP_TILDE 4
#define OP_EXCLEXCL 5
#define OP_PERCEQUAL 6
#define OP_ANDAND 7
#define OP_STAREQUAL 8
#define OP_PLUSPLUS 9
#define OP_MINUSMINUS 10
#define OP_LESSEQUAL 11
#define OP_EQUALEQUAL 12
#define OP_GTEQUAL 13
#define OP_CARETEQUAL 14
#define OP_BARBAR 15
#define OP_EXCLEQUAL 16
#define OP_EXCL 17
#define OP_PERCENT 18
#define OP_ANDANDEQUAL 19
#define OP_ANDEQUAL 20
#define OP_AND 21
#define OP_ASTERISK 22
#define OP_PLUSEQUAL 23
#define OP_PLUS 24
#define OP_MINUSEQUAL 25
#define OP_MINUSGT 26
#define OP_MINUS 27
#define OP_SLASHEQUAL 28
#define OP_SLASH 29
#define OP_LESS 30
#define OP_LESSLESSEQUAL 31
#define OP_LESSLESS 32
#define OP_EQUAL 33
#define OP_GT 34
#define OP_GTGTEQUAL 35
#define OP_GTGT 36
#define OP_CARET 37
#define OP_BARBAREQUAL 38
#define OP_BAREQUAL 39
#define OP_BAR 40
#define OP_HASH 41
#define OP_HASHHASH 42

/* Special codes */

#define SPEC_LPAREN 8
#define SPEC_RPAREN 9
#define SPEC_SEMIC 11
#define SPEC_LBRACK 13
#define SPEC_RBRACK 14
#define SPEC_BACKSLASH 15
#define SPEC_LGROUP 16
#define SPEC_RGROUP 17
