/*
List Handling for C-Source TeX Formatter
(c) 1995 Martin Mares \& Robert Spalek
*/

#include "cstf.h"

struct listnode firstnode,lastnode;

union stringrec {
	union stringrec *next;
	byte string[256];
	};

static struct listnode *firstfreen;
static union stringrec *firstfrees;

void init_list(void)
{
	firstnode.next = &lastnode;
	lastnode.prev = &firstnode;
}

static struct listnode *new_node(int flag)
{
struct listnode *r;
	if (firstfreen)
		{
		r=firstfreen;
		firstfreen = r->next;
		}
	else r=xmalloc(sizeof(struct listnode));
	r->next = NULL;
	if ((r->flag = flag) == 1)
		if (firstfrees)
			{
			LVAL(r) = (unsigned int) firstfrees->string;
			firstfrees = firstfrees->next;
			}
		else
			LVAL(r) = (unsigned int) xmalloc(256);
	return r;
}

static void delete_node(struct listnode *r)
{
	r->prev->next = r->next;
	r->next->prev = r->prev;
	switch (r->flag)
		{
		case 1:
			{
			union stringrec *s = (union stringrec *) r->val;
				s->next = firstfrees;
				firstfrees = s;
				break;
			}
		case 2:
			free(LSTR(r));
			break;
		}
	r->next = firstfreen;
	firstfreen = r;
}

void insert_node(struct listnode *r,struct listnode *after)
{
	r->next = after->next;
	r->prev = after;
	after->next = r;
	r->next->prev = r;
}

void append_node(struct listnode *r)
{
	r->prev = lastnode.prev;
	r->prev->next = r;
	r->next = &lastnode;
	lastnode.prev = r;
}

#if 1

static byte opernames[] = ".\000\000\000,\000\000\000:\000\000\000?\000\000\000\
~\000\000\000!\000\000\000%=\000\000&&\000\000*=\000\000++\000\000--\000\000\
<=\000\000==\000\000>=\000\000&=\000\000||\000\000!=\000\000!\000\000\000\
%\000\000\000&&=\000&=\000\000&\000\000\000*\000\000\000+=\000\000+\000\000\000\
-=\000\000->\000\000-\000\000\000/=\000\000/\000\000\000<\000\000\000<<=\000\
<<\000\000=\000\000\000>\000\000\000>>=\000>>\000\000^\000\000\000||=\000\
|=\000\000|\000\000\000#\000\000\000##\000\000";

static byte specnames[] = "@@@@@@@@()@;@[]@{}";

void debug_list(void)
{
struct listnode *prev=&firstnode, *r=firstnode.next;
	while (r->next)
		{
		if (r->prev != prev) Err("INTERNAL ERROR #2");
		switch (r->cate)
			{
			case L_EOF:
				printf("EOF "); break;
			case L_EOL:
				printf("EOL "); break;

			case L_COMMENT:
				printf("COMMENT(%s) ",LSTR(r)); break;
			case L_FULLCOMMENT:
				printf("FULLCOMMENT(%s) ",LSTR(r)); break;
			case L_CHAR:
				printf("CHAR(%s) ",LSTR(r)); break;
			case L_NUMBER:
				printf("NUM(%s) ",LSTR(r)); break;
			case L_STRING:
				printf("STR(%s) ",LSTR(r)); break;
			case L_INCLARG:
				printf("INCL(%s) ",LSTR(r)); break;
			case L_PREPROC:
				printf("DIR(%s) ",LSTR(r)); break;
			case L_COMMAND:
				printf("CMD(%s) ",LSTR(r)); break;
			case L_TYPE:
				printf("TYPE(%s) ",LSTR(r)); break;
			case L_VAR:
				printf("VAR(%s) ",LSTR(r)); break;
			case L_WORD:
				printf("WORD(%s) ",LSTR(r)); break;
			case L_SPLITCOMMENT:
				printf("SPLITCOMMENT(%s) ",LSTR(r)); break;
			case L_STORCLASS:
				printf("CLASS(%s) ",LSTR(r)); break;
			case L_FUNCTION:
				printf("FUNC(%s) ",LSTR(r)); break;

			case L_OPERATOR:
				printf("OPER(%s) ",opernames + 4*LVAL(r)); break;
			case L_SPECIAL:
				printf("OPER(%c) ",specnames[LVAL(r)]); break;
			case L_SPACE:
				printf("SPACE(%d) ",LVAL(r)); break;
			case L_COMMENTFILL:
				printf("CFILL(%d) ",LVAL(r)); break;

			default:
				printf("???(%d) ",LVAL(r));
			}
		prev = r;
		r=r->next;
		}
	if (r != &lastnode || firstnode.prev) Err("INTERNAL ERROR #1");
	putchar('\n');
}

#endif

void delete_list(void)
{
struct listnode *s,*r = firstnode.next;
	while (s = r->next)
		{
		delete_node(r);
		r=s;
		}
}

void insert_new_node(struct listnode *after,int cate,int val)
{
struct listnode *t = new_node(0);
	t->cate = cate; LVAL(t) = val;
	insert_node(t,after);
}

static int already_started;

void create_list(void)
{
struct listnode *r;
	do
		{
		if (!already_started)
			{
			class=L_EOL;	/* Add fictive EOL at the start */
			already_started = 1;
			}
		else parse();
		if (class == L_EOF) break;
		switch (class)
			{
			case L_COMMENT:
			case L_FULLCOMMENT:
			case L_CHAR:
			case L_NUMBER:
			case L_STRING:
			case L_INCLARG:
			case L_PREPROC:
			case L_COMMAND:
			case L_TYPE:
			case L_VAR:
			case L_WORD:
			case L_SPLITCOMMENT:
			case L_STORCLASS:
			case L_FUNCTION:
				{
				int l;
					if ((l=strlen(strbuf)) > 255)
						{
						r = new_node(2);
						r->val = (unsigned int) xmalloc(l+1);
						}
					else r = new_node(1);
					strcpy(LSTR(r),strbuf);
					break;
				}

			default:
				r = new_node(0);
				LVAL(r) = value;
			}
		r->cate = class;
		append_node(r);
		}
	while (class != L_SPECIAL || value != SPEC_SEMIC);
}

static void eol_correct(struct listnode *r)
{
	if (r->next->cate != L_SPACE && r->next->cate != L_EOL &&
		r->next->cate != L_FULLCOMMENT)
			insert_new_node(r,L_SPACE,-1);
}

/* Changes:
%*
%* ~EOL SPACE *COMMENT  ->  *COMMENT
%* *COMMENT SPACE  ->  *COMMENT
%* WORD #(SPACE | COMMENT) SPEC(()  ->  FUNCTION ...
%* FULLCOMMENT ~EOL   ->  SPLITCOMMENT *
%* ~EOL (COMMENT | SPLITCOMMENT)  ->  * COMMENTFILL(0) *
%* (COMMENT | SPLITCOMMENT) ~(COMMENT | SPLITCOMMENT | EOL)  -> * COMMENTFILL(1) *
%* COMMENTFILL(0) (COMMENT | SPLITCOMMENT) ~COMMENTFILL  ->  COMMENTFILL(2) * *
%* EOL #EOL FULLCOMMENT  ->  EOL FULLCOMMENT
%* FULLCOMMENT #EOL  ->  FULLCOMMENT
%* (EOL | FULLCOMMENT) ~SPACE  ->  EOL SPACE(-1)
*/

void process_list(void)
{
struct listnode *r,*s;
	r = firstnode.next;
	s = NULL;
	while (r->next)
		{
		switch (r->cate)
			{
			case L_COMMENT:
			case L_SPLITCOMMENT:
			case L_FULLCOMMENT:
				if (r->prev->cate == L_SPACE && r->prev->prev->cate != L_EOL)
					delete_node(r->prev);
				if (r->next->cate == L_SPACE)
					delete_node(r->next);
				if (r->cate == L_FULLCOMMENT)
					if (r->next->cate != L_EOL)
						r->cate = L_SPLITCOMMENT;
					else
						{
						while (r->next->cate == L_EOL)
							delete_node(r->next);
						while (r->prev->cate == L_EOL && r->prev->prev->cate == L_EOL)
							delete_node(r->prev);
						eol_correct(r);
						}
				if (r->cate == L_COMMENT || r->cate == L_SPLITCOMMENT)
					{
					int i;
						if (r->prev->cate != L_EOL && r->prev->cate != L_SPACE)
							insert_new_node(r->prev,L_COMMENTFILL,0);
						i = r->next->cate;
						if (i != L_COMMENT && i != L_SPLITCOMMENT && i != L_EOL)
							insert_new_node(r,L_COMMENTFILL,1);
						if (r->prev->cate == L_COMMENTFILL && LVAL(r->prev) == 0
							&& r->next->cate != L_COMMENTFILL)
								LVAL(r->prev) = 2;
					}
				/* Fall-thru */
			case L_SPACE:
				break;

			case L_SPECIAL:
				if (LVAL(r) == SPEC_LPAREN && s) s->cate = L_FUNCTION;
				s = NULL;
				break;

			case L_WORD:
				s = r;
				break;

			case L_EOL:
				eol_correct(r);
				break;

			default:
				s = NULL;
			}
		r = r->next;
		}
	if (already_started == 1)
		{
		while (firstnode.next->cate == L_EOL) delete_node(firstnode.next);
		already_started = 2;
		}
}
