/*
Options Processing for the C-Source TeX Formatter
(c) 1995 Martin Mares \& Robert Spalek
*/

#include "cstf.h"

int TabSize=8;
int StringMode=1;
int InsertBye=0;
int InsertMacros=1;
int debug_value=0;

static void OptErr(byte *e,byte *o)
{
	fprintf(stderr,"Option `%s' in ",o);
	if (line) fprintf(stderr,"line %d",line);
	else fprintf(stderr,"command line");
	fprintf(stderr,": %s !\n",e);
	exit(FAIL_RETURN_CODE);
}

struct optrec {
	byte *name;
	byte arg;
	};

#define A_NONE 0
#define A_NUM 1
#define A_STR 2

static struct optrec opttab[] = {
	{ "tabsize", A_NUM },
	{ "ssv", A_NONE },
	{ "ssi", A_NONE },
	{ "ssa", A_NONE },
	{ "bye", A_NONE },
	{ "nomac", A_NONE },
	{ "cmd", A_STR },
	{ "type", A_STR },
	{ "func", A_STR },
	{ "class", A_STR },
	{ "debug", A_NUM },
	{ "var", A_STR }
	};

static void DefNewKwd(byte *str,int cate)
{
struct symrec *r = new_symbol(str);
	r->cate = cate;
}

void process_option(byte *opt)
{
byte *c = opt;
byte t,*str;
int i,num;
	str = NULL;
	num = 0;
	while (*c && *c != '=') c++;
	t=*c; *c=0;
	for (i=0;i<(sizeof(opttab)/sizeof(struct optrec));i++)
		if (!strcmp(opt,opttab[i].name))
			{
			*c=t;
			if (opttab[i].arg)
				{
				if (!t)
					OptErr("Missing option argument",opt);
				c++;
				if (opttab[i].arg == A_STR) str=c;
				else num=atol(c);
				}
			else if (t)
				OptErr("Unexpected argument",opt);
			switch (i)
				{
				case 0:
					if (num < 1 || num > 32)
						OptErr("Invalid tab size",opt);
					TabSize=num;
					break;
				case 1:
					StringMode=2;
					break;
				case 2:
					StringMode=0;
					break;
				case 3:
					StringMode=1;
					break;
				case 4:
					InsertBye=1;
					break;
				case 5:
					InsertMacros=0;
					break;
				case 6:
					DefNewKwd(str,L_COMMAND);
					break;
				case 7:
					DefNewKwd(str,L_TYPE);
					break;
				case 8:
					DefNewKwd(str,L_FUNCTION);
					break;
				case 9:
					DefNewKwd(str,L_STORCLASS);
					break;
				case 10:
					debug_value = num;
					break;
				case 11:
					DefNewKwd(str,L_VAR);
					break;
				}
			return;
			}
	*c=t;
	OptErr("Invalid option",opt);
}
