/*
TeX Output for C-Source TeX Formatter
(c) 1995 Martin Mares \& Robert Spalek
*/

#include "cstf.h"

#define OUTBUFSIZE 256
#define OUTLIMIT1 70
#define OUTLIMIT2 80

static byte out_buf[OUTBUFSIZE];
static byte format_buf[256];
static int out_index;

void init_tex(void)
{
	fprintf(outfile,"%% CSTF output. Don't edit manually if it's possible to do it \
by modifying the C source.\n");
	if (InsertMacros)
		fprintf(outfile,"\\input cstfmac.tex\n");
	fputc('\n',outfile);
}

void end_tex(void)
{
	if (InsertBye) fprintf(outfile,"\\bye\n");
}

static void flush_out(int n)
{
	if (out_index)
		{
		fwrite(out_buf,out_index,1,outfile);
		if (n) fputc('%',outfile);
		fputc('\n',outfile);
		out_index = 0;
		}
}

static void split_on(byte c)
{
int l = out_index;
int m = out_index;
	while (l && out_buf[l-1] != c) l--;
	if (l)
		{
		out_index = --l;
		flush_out(c != ' ');
		if (c == ' ') l++;
		memcpy(out_buf,out_buf+l,m-l);
		out_index = m-l;
		}
}

static void outchar(byte c)
{
	if (c == '\n')
		{
		if (out_index) flush_out(0);
		else fputc('\n',outfile);
		}
	else
		{
		if (c != ' ' || (out_index && out_buf[out_index-1] != ' '))
			out_buf[out_index++] = c;
		if (out_index >= OUTLIMIT1) split_on(' ');
		if (out_index >= OUTLIMIT2) split_on('\\');
		if (out_index >= OUTBUFSIZE)
			Err("Internal buffer overflow");
		}
}

static void outstr(byte *c)
{
	while (*c) outchar(*c++);
}

static void output(byte *c,byte *s)
{
	sprintf(format_buf,c,s);
	outstr(format_buf);
}

static void outnum(byte *c,int l)
{
	sprintf(format_buf,c,l);
	outstr(format_buf);
}

static byte *specials[] = { NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
"Lparen","Rparen",NULL,"Semic",NULL,"Lbrack","Rbrack","Backslash","Lbrace",
"Rbrace" };

static byte *opers[] = { "Dot", "Comma", "Colon", "Quest", "Tilde", "Exclexcl",
"Percequal", "Andand", "Starequal", "Plusplus", "Minusminus", "Lessequal",
"Equalequal", "Gtequal", "Caretequal", "Barbar", "Exclequal", "Excl", "Percent",
"Andandequal", "Andequal", "And", "Asterisk", "Plusequal", "Plus", "Minusequal",
"Minusgt", "Minus", "Slashequal", "Slash", "Less", "Lesslessequal", "Lessless",
"Equal", "Gt", "Gtgtequal", "Gtgt", "Caret", "Barbarequal", "Barequal", "Bar",
"Hash", "Hashhash" };

static void reduce_comment(struct listnode *r)
{
byte *k,*z = k = LSTR(r);
	while (*z) z++;
	while (z != k && z[-1] == '\n') z--;
	while (*k == '\n') k++;
	while (k < z) outchar(*k++);
}

static void emit_spaces(int l,int flag)
{
	if (l == -1) outstr("\\bl");
	else if (!l) outchar(' ');
	else
		{
		if (flag) outstr("\\sp");
		else outstr("\\spc");
		if (l < 10) outnum("%d ",l);
		else outnum("{%d} ",l);
		}
}

static void emit_string(byte *cs,byte *t)
{
byte *z,*old;
	output("\\%s ",cs);
	while (*t == ' ')
		{
		outstr("\\vs ");
		t++;
		}
	z=NULL;
	while (*t)
		{
		old=z; z=NULL;
		switch (*t++)
			{
			case '`': z="bt"; break;
			case '~': z="ti"; break;
			case '@': z="at"; break;
			case '#': z="ha"; break;
			case '$': z="do"; break;
			case '%': z="pe"; break;
			case '^': z="ca"; break;
			case '&': z="an"; break;
			case '_': z="un"; break;
			case '-': z="mi"; break;
			case '\\': z="ba"; break;
			case '|': z="ve"; break;
			case '{': z="lg"; break;
			case '}': z="rg"; break;
			case '[': z="lb"; break;
			case ']': z="rb"; break;
			case '\'': z="ap"; break;
			case '"': z="qo"; break;
			case '?': z="qu"; break;
			case '>': z="gt"; break;
			case '<': z="lt"; break;
			case '*': z="as"; break;
			case '\n': outstr("\\eol\n\\bl"); break;
			case ' ':
				{
				byte *s = --t;
					while (*t == ' ') t++;
					if (StringMode == 2 || StringMode == 1 && (t > s+1 || !*t))
						while (s < t)
							{
							s++;
							outstr("\\vs");
							}
					else if (t-s > 1) emit_spaces(t-s,0);
					else
						{
						if (old) outchar('\\');
						outchar(' ');
						}
					break;
				}
			}
		if (z) output("\\S%s ",z);
		else outchar(t[-1]);
		}
	output("\\e%s ",cs);
}

void output_tex(void)
{
int sp;
struct listnode *r;

	for (r = firstnode.next ; r->next ; r=r->next)
		{
		sp = 0;
		switch (r->cate)
			{
			case L_COMMENT:
			case L_SPLITCOMMENT:
				outstr("{\\comment ");
				outstr(LSTR(r));
				outchar('}');
				break;
			case L_FULLCOMMENT:
				outstr("\\begincomment\n");
				reduce_comment(r);
				outstr("\n\\endcomment\n");
				break;
			case L_OPERATOR:
				output("\\%s",opers[LVAL(r)]);
				sp=1; break;
			case L_NUMBER:
				{
				byte *c = LSTR(r);
					if (*c == '0' && (c[1] & 0xdf) == 'X')
						output("\\hex{%s}",c+2);
					else
						{
						outchar('$');
						while (*c)
							{
							if (*c == 'e' || *c == 'E')
								{
								outchar('\\'); outchar(*c); outchar(' ');
								}
							else if (*c == '+') outstr("\\rplus ");
							else if (*c == '-') outstr("\\rminus ");
							else outchar(*c);
							c++;
							}
						outchar('$');
						}
					break;
				}
			case L_CHAR:
				emit_string("Char",LSTR(r));
				break;
			case L_STRING:
				emit_string("String",LSTR(r));
				break;
			case L_EOL:
				if (r->prev->cate == L_EOL) outstr("\\empty\n");
				else outstr("\\eol\n");
				break;
			case L_INCLARG:
				output("\\inclarg{%s}",LSTR(r)); break;
			case L_PREPROC:
				output("\\preproc{%s}",LSTR(r)); break;
			case L_COMMAND:
				output("\\cmd{%s}",LSTR(r)); break;
			case L_TYPE:
				output("\\type{%s}",LSTR(r)); break;
			case L_WORD:
			case L_VAR:
				output("\\var{%s}",LSTR(r)); break;
			case L_SPECIAL:
				output("\\%s",specials[LVAL(r)]);
				sp=1; break;
			case L_SPACE:
				emit_spaces(LVAL(r),1);
				break;
			case L_STORCLASS:
				output("\\class{%s}",LSTR(r)); break;
			case L_FUNCTION:
				output("\\func{%s}",LSTR(r)); break;
			case L_COMMENTFILL:
				switch (LVAL(r))
					{
					case 0: outstr("\\precf"); break;
					case 1: outstr("\\postcf"); break;
					case 2: outstr("\\preci"); break;
					}
				break;
			}
		if (sp && r->next->cate == L_SPACE) outstr("{}");
		}
}
